"""Read file tool"""
from typing import Any, Dict
from .base import BaseTool


class ReadFileTool(BaseTool):
    """Read content from files"""
    
    @property
    def name(self) -> str:
        return "read_file"
    
    @property
    def description(self) -> str:
        return "Read content from a file"
    
    @property
    def parameters(self) -> Dict[str, Any]:
        return {
            "type": "object",
            "properties": {
                "file_path": {
                    "type": "string",
                    "description": "Path to the file to read"
                }
            },
            "required": ["file_path"]
        }
    
    def execute(self, file_path: str) -> Dict[str, Any]:
        """Read file content"""
        try:
            with open(file_path, 'r', encoding='utf-8') as f:
                content = f.read()
            return {
                "success": True,
                "content": content
            }
        except Exception as e:
            return {
                "success": False,
                "error": str(e)
            }



"""Execute command tool with interactive shell sessions"""
import os
import subprocess
import platform
import threading
import time
import uuid
from collections import deque
from typing import Any, Dict, Optional, Deque, Tuple, List
from .base import BaseTool, BaseToolGroup, BaseStatus
from ..config import settings


class ShellSession:
    """Persistent shell session with async output capture"""

    def __init__(self, working_dir: Optional[str] = None):
        self.working_dir = working_dir
        self._output: Deque[Tuple[str, str]] = deque()
        self._log: Deque[Dict[str, str]] = deque()
        self._output_event = threading.Event()
        self._lock = threading.Lock()
        self._running_marker: Optional[str] = None  # Track if command is running

        env = os.environ.copy()
        env["PYTHONIOENCODING"] = "utf-8"
        
        # Use UTF-8 encoding for both Windows and Unix
        # Git Bash and Python both support UTF-8 well
        encoding = "utf-8"
        
        # Disable colors and set simple terminal to avoid ANSI escape sequences
        env["TERM"] = "dumb"
        env["PS1"] = "$ "  # Simple prompt without colors
        env["NO_COLOR"] = "1"  # Disable colors in many tools
        env["PYTHONUNBUFFERED"] = "1"  # Disable Python output buffering
        
        # Always use bash, get path from config
        bash_cmd = settings.bash_path
        # Use --norc --noprofile to avoid loading user configs that add colors
        # Don't use -i to avoid "no job control" warnings in non-TTY environment
        cmd = [bash_cmd, "--norc", "--noprofile"]

        self.process = subprocess.Popen(
            cmd,
            stdin=subprocess.PIPE,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            cwd=working_dir if working_dir else None,
            text=True,
            bufsize=1,
            env=env,
            encoding=encoding,
            errors="replace",
        )

        self._stdout_thread = threading.Thread(
            target=self._read_stream,
            args=("stdout", self.process.stdout),
            daemon=True,
        )
        self._stderr_thread = threading.Thread(
            target=self._read_stream,
            args=("stderr", self.process.stderr),
            daemon=True,
        )
        self._stdout_thread.start()
        self._stderr_thread.start()

    def _read_stream(self, stream_name: str, stream):
        try:
            for line in iter(stream.readline, ""):
                if line:
                    with self._lock:
                        self._output.append((stream_name, line))
                        self._log.append({"stream": stream_name, "data": line})
                        self._output_event.set()
        except Exception:
            pass
        finally:
            stream.close()

    def is_alive(self) -> bool:
        return self.process and self.process.poll() is None

    def write(self, data: str) -> None:
        if not self.is_alive():
            raise RuntimeError("Shell session is not running")
        if not data.endswith("\n"):
            data += "\n"
        with self._lock:
            self._log.append({"stream": "stdin", "data": data})
        self.process.stdin.write(data)
        self.process.stdin.flush()

    def read(self, wait_ms: int = 0, max_chars: int = 20000) -> Dict[str, Any]:
        output = []
        stderr = []

        if wait_ms > 0:
            self._output_event.wait(wait_ms / 1000)

        with self._lock:
            char_count = 0
            while self._output and char_count <= max_chars:
                stream_name, line = self._output.popleft()
                char_count += len(line)
                if stream_name == "stdout":
                    output.append(line)
                else:
                    stderr.append(line)
            self._output_event.clear()

        return {
            "stdout": "".join(output),
            "stderr": "".join(stderr),
        }

    def terminate(self) -> None:
        if self.is_alive():
            self.process.terminate()

    def get_log(self) -> list:
        with self._lock:
            return list(self._log)
    
    def is_running(self) -> bool:
        """Check if a command is currently running (marker not yet seen)"""
        if self._running_marker is None:
            return False
        
        # Check if marker is already in the buffered output
        # This handles the case where command completed but we didn't wait long enough
        with self._lock:
            new_output = deque()
            marker_found = False
            for stream_name, line in self._output:
                if self._running_marker in line:
                    # Marker found in buffer, command has completed
                    marker_found = True
                    # Remove the marker line from output
                    continue
                new_output.append((stream_name, line))
            
            if marker_found:
                self._output = new_output
                self._running_marker = None
                return False
        
        return True
    
    def set_running_marker(self, marker: str) -> None:
        """Set the marker for currently running command"""
        with self._lock:
            self._running_marker = marker
    
    def clear_running_marker(self) -> None:
        """Clear the running marker when command completes"""
        with self._lock:
            self._running_marker = None


_shell_sessions: Dict[str, ShellSession] = {}


def _collect_output_until_marker(
    session: ShellSession,
    marker: str,
    wait_ms: int,
    max_chars: int,
) -> Tuple[str, str, bool]:
    """
    Collect output from session until marker is found or timeout.
    
    Returns: (stdout, stderr, marker_found)
    """
    max_wait_time = wait_ms if wait_ms > 0 else 100000
    start_time = time.time() * 1000
    all_stdout = []
    all_stderr = []
    marker_found = False

    while (time.time() * 1000 - start_time) < max_wait_time:
        output = session.read(wait_ms=100, max_chars=max_chars)
        stdout_chunk = output.get("stdout", "")
        stderr_chunk = output.get("stderr", "")

        if marker in stdout_chunk:
            # Remove the marker and everything after it (including the marker line)
            lines = stdout_chunk.split('\n')
            filtered_lines = []
            for line in lines:
                if marker in line:
                    marker_found = True
                    break
                filtered_lines.append(line)
            stdout_chunk = '\n'.join(filtered_lines)
            # Add final newline if there were lines
            if filtered_lines and stdout_chunk and not stdout_chunk.endswith('\n'):
                stdout_chunk += '\n'

        if stdout_chunk:
            all_stdout.append(stdout_chunk)
        if stderr_chunk:
            all_stderr.append(stderr_chunk)

        if marker_found:
            # Clear running marker when command completes
            session.clear_running_marker()
            break

        if not stdout_chunk and not stderr_chunk:
            time.sleep(0.01)

    return "".join(all_stdout), "".join(all_stderr), marker_found


def _ensure_session(session_id: str, working_dir: Optional[str] = None) -> ShellSession:
    if session_id not in _shell_sessions or not _shell_sessions[session_id].is_alive():
        _shell_sessions[session_id] = ShellSession(working_dir=working_dir)
        if working_dir:
            _shell_sessions[session_id].write(f'cd "{working_dir}"')
    return _shell_sessions[session_id]


def _get_sessions_status() -> List[Dict[str, Any]]:
    sessions = []
    for session_id, session in _shell_sessions.items():
        sessions.append({
            "session_id": session_id,
            "working_dir": session.working_dir,
            "alive": session.is_alive(),
            "pid": session.process.pid if session.process else None,
            "log": session.get_log(),
        })
    return sessions


class ShellStartTool(BaseTool):
    """Start a persistent shell session"""

    @property
    def name(self) -> str:
        return "shell_start"

    @property
    def description(self) -> str:
        return "Start a persistent bash shell session."

    @property
    def parameters(self) -> Dict[str, Any]:
        return {
            "type": "object",
            "properties": {
                "session_id": {
                    "type": "string",
                    "description": "Session/agent identifier for persistent shell",
                },
                "working_dir": {
                    "type": "string",
                    "description": "Working directory to start shell (optional)",
                },
            },
            "required": [],
        }

    def execute(
        self,
        session_id: Optional[str] = None,
        working_dir: Optional[str] = None,
    ) -> Dict[str, Any]:
        session_id = session_id or "default"
        _ensure_session(session_id, working_dir=working_dir)
        return {
            "success": True,
            "session_id": session_id,
            "status": "started",
        }


class ShellRunTool(BaseTool):
    """Run a command in a shell session"""

    @property
    def name(self) -> str:
        return "shell_run"

    @property
    def description(self) -> str:
        return "Run a command in a persistent bash session. **IMPORTANT**: If 'background' is set to true, the command will run in the background and the tool will return immediately. This will occupy the shell session; start a new session for other commands."

    @property
    def parameters(self) -> Dict[str, Any]:
        return {
            "type": "object",
            "properties": {
                "session_id": {
                    "type": "string",
                    "description": "Session/agent identifier for persistent shell",
                },
                "command": {
                    "type": "string",
                    "description": "Command to run",
                },
                "wait_ms": {
                    "type": "integer",
                    "description": "Max wait time in milliseconds for command completion",
                },
                "max_chars": {
                    "type": "integer",
                    "description": "Max characters to return from buffered output",
                    "default": 20000,
                },
                "background": {
                    "type": "boolean",
                    "description": "Run command in background and return immediately. **IMPORTANT** This will occupy the shell session; start a new session for other commands.",
                    "default": False,
                },
                "working_dir": {
                    "type": "string",
                    "description": "Working directory to start shell (optional)",
                },
            },
            "required": ["command", "background"],
        }

    def execute(
        self,
        command: str,
        session_id: Optional[str] = None,
        wait_ms: Optional[int] = None,
        max_chars: int = 20000,
        background: bool = False,
        working_dir: Optional[str] = None,
    ) -> Dict[str, Any]:
        session_id = session_id or "default"
        session = _ensure_session(session_id, working_dir=working_dir)

        # Check if session is already running a command
        if session.is_running():
            return {
                "success": False,
                "error": f"Session '{session_id}' is already running a command. Please wait for it to complete, use shell_read to check status, or start a new session to run another command.",
                "session_id": session_id,
            }

        if wait_ms is None:
            wait_ms = 10000 if background else 100000

        marker = f"__CMD_DONE_{uuid.uuid4().hex[:8]}__"
        
        # Mark session as running before executing command
        session.set_running_marker(marker)

        session.write(command)
        session.write(f"echo {marker}")

        if background:
            # For background mode, wait for marker with a reasonable timeout
            # If marker is found, command completed and marker will be cleared
            # If marker is not found, command is still running, keep marker set
            # (is_running() will check buffer for marker before rejecting new commands)
            stdout, stderr, marker_found = _collect_output_until_marker(
                session, marker, wait_ms, max_chars
            )
            
            status = "completed" if marker_found else "running"
            
            return {
                "success": True,
                "session_id": session_id,
                "status": status,
                "stdout": stdout,
                "stderr": stderr,
            }
        else:
            # For normal mode, wait until completion or timeout
            max_wait_time = wait_ms if wait_ms > 0 else 100000
            stdout, stderr, marker_found = _collect_output_until_marker(
                session, marker, max_wait_time, max_chars
            )
            
            return {
                "success": True,
                "session_id": session_id,
                "stdout": stdout,
                "stderr": stderr,
            }


class ShellWriteTool(BaseTool):
    """Write input to a shell session"""

    @property
    def name(self) -> str:
        return "shell_write"

    @property
    def description(self) -> str:
        return "Write input to stdin of a persistent bash session."

    @property
    def parameters(self) -> Dict[str, Any]:
        return {
            "type": "object",
            "properties": {
                "session_id": {
                    "type": "string",
                    "description": "Session/agent identifier for persistent shell",
                },
                "input": {
                    "type": "string",
                    "description": "Input to write to stdin",
                },
                "working_dir": {
                    "type": "string",
                    "description": "Working directory to start shell (optional)",
                },
            },
            "required": ["input"],
        }

    def execute(
        self,
        input: str,
        session_id: Optional[str] = None,
        working_dir: Optional[str] = None,
    ) -> Dict[str, Any]:
        session_id = session_id or "default"
        session = _ensure_session(session_id, working_dir=working_dir)
        session.write(input)
        return {
            "success": True,
            "session_id": session_id,
        }


class ShellReadTool(BaseTool):
    """Read output from a shell session"""

    @property
    def name(self) -> str:
        return "shell_read"

    @property
    def description(self) -> str:
        return "Read buffered output from a persistent bash session."

    @property
    def parameters(self) -> Dict[str, Any]:
        return {
            "type": "object",
            "properties": {
                "session_id": {
                    "type": "string",
                    "description": "Session/agent identifier for persistent shell",
                },
                "wait_ms": {
                    "type": "integer",
                    "description": "Wait time in milliseconds before reading output",
                    "default": 0,
                },
                "max_chars": {
                    "type": "integer",
                    "description": "Max characters to return from buffered output",
                    "default": 20000,
                },
                "working_dir": {
                    "type": "string",
                    "description": "Working directory to start shell (optional)",
                },
            },
            "required": [],
        }

    def execute(
        self,
        session_id: Optional[str] = None,
        wait_ms: int = 0,
        max_chars: int = 20000,
        working_dir: Optional[str] = None,
    ) -> Dict[str, Any]:
        session_id = session_id or "default"
        session = _ensure_session(session_id, working_dir=working_dir)
        default_wait_ms = wait_ms if wait_ms > 0 else 1000
        output = session.read(wait_ms=default_wait_ms, max_chars=max_chars)
        return {
            "success": True,
            "session_id": session_id,
            **output,
        }


class ShellStopTool(BaseTool):
    """Stop a shell session"""

    @property
    def name(self) -> str:
        return "shell_stop"

    @property
    def description(self) -> str:
        return "Stop a persistent bash session."

    @property
    def parameters(self) -> Dict[str, Any]:
        return {
            "type": "object",
            "properties": {
                "session_id": {
                    "type": "string",
                    "description": "Session/agent identifier for persistent shell",
                },
            },
            "required": [],
        }

    def execute(self, session_id: Optional[str] = None) -> Dict[str, Any]:
        session_id = session_id or "default"
        session = _shell_sessions.get(session_id)
        if not session or not session.is_alive():
            return {
                "success": True,
                "session_id": session_id,
                "status": "stopped",
            }
        session.terminate()
        return {
            "success": True,
            "session_id": session_id,
            "status": "stopped",
        }


class ShellToolGroup(BaseToolGroup):
    """Shell tool group"""

    @property
    def name(self) -> str:
        return "shell"

    @property
    def description(self) -> str:
        return "Persistent bash shell tools."

    def get_tools(self) -> List[BaseTool]:
        return [
            ShellStartTool(),
            ShellRunTool(),
            ShellWriteTool(),
            ShellReadTool(),
            ShellStopTool(),
        ]


class ShellStatus(BaseStatus):
    """Status provider for shell sessions"""

    @property
    def name(self) -> str:
        return "shell"

    @property
    def description(self) -> str:
        return "Persistent bash shell status."

    def get_status(self) -> Dict[str, Any]:
        return {
            "name": self.name,
            "status": "ok",
            "sessions": _get_sessions_status(),
        }



"""Command-line interface for running the IriBot backend."""

from __future__ import annotations

import argparse
import uvicorn


def main() -> None:
    parser = argparse.ArgumentParser(description="Run the IriBot backend server")
    parser.add_argument("--host", default="127.0.0.1", help="Host to bind")
    parser.add_argument("--port", type=int, default=8000, help="Port to bind")
    parser.add_argument("--reload", action="store_true", help="Enable auto-reload")
    parser.add_argument(
        "--log-level",
        default="info",
        choices=["critical", "error", "warning", "info", "debug", "trace"],
        help="Log level",
    )

    args = parser.parse_args()

    uvicorn.run(
        "iribot.main:app",
        host=args.host,
        port=args.port,
        reload=args.reload,
        log_level=args.log_level,
    )


if __name__ == "__main__":
    main()



"""Core Agent logic for handling LLM interactions"""

import json
from typing import Optional, List, Dict, Any, Generator
from openai import OpenAI
from .config import settings
from .executor import tool_executor


class Agent:
    """Agent for handling LLM interactions"""

    def __init__(self):
        client_params = {"api_key": settings.openai_api_key}
        if settings.openai_base_url:
            client_params["base_url"] = settings.openai_base_url

        self.client = OpenAI(**client_params)
        self.model = settings.openai_model

    def chat_stream(
        self,
        messages: List[Dict[str, Any]],
        system_prompt: str,
        images: Optional[List[str]] = None,
    ) -> Generator[Dict[str, Any], None, None]:
        """
        Send a message to the LLM and stream the response

        Yields:
            Chunks of response data
        """
        # Build messages with system prompt
        formatted_messages = [{"role": "system", "content": system_prompt}]

        # Add image to the last user message if provided
        if images and messages and messages[-1]["role"] == "user":
            last_msg = messages[-1].copy()
            content = [{"type": "text", "text": last_msg.get("content", "")}]

            for image_base64 in images:
                content.append(
                    {
                        "type": "image_url",
                        "image_url": {
                            "url": f"data:image/jpeg;base64,{image_base64}"
                        },
                    }
                )

            last_msg["content"] = content
            formatted_messages.extend(messages[:-1])
            formatted_messages.append(last_msg)
        else:
            formatted_messages.extend(messages)

        # Get available tools
        tools = tool_executor.get_all_tools()

        # Call OpenAI API with streaming
        response = self.client.chat.completions.create(
            model=self.model,
            messages=formatted_messages,
            tools=tools,
            temperature=0.7,
            stream=True,
        )

        content = ""
        tool_calls_data = {}  # {index: {id, function: {name, arguments}}}
        finish_reason = None

        for chunk in response:
            delta = chunk.choices[0].delta if chunk.choices else None

            if delta:
                # Handle content
                if delta.content:
                    content += delta.content
                    yield {"type": "content", "content": delta.content}

                # Handle tool calls
                if delta.tool_calls:
                    for tc in delta.tool_calls:
                        idx = tc.index
                        if idx not in tool_calls_data:
                            tool_calls_data[idx] = {
                                "id": tc.id or "",
                                "type": "function",
                                "function": {"name": "", "arguments": ""},
                            }

                        if tc.id:
                            tool_calls_data[idx]["id"] = tc.id
                        if tc.function:
                            if tc.function.name:
                                tool_calls_data[idx]["function"][
                                    "name"
                                ] = tc.function.name
                            if tc.function.arguments:
                                tool_calls_data[idx]["function"][
                                    "arguments"
                                ] += tc.function.arguments

            if chunk.choices and chunk.choices[0].finish_reason:
                finish_reason = chunk.choices[0].finish_reason

        # Yield final result with tool calls
        tool_calls = (
            [tool_calls_data[i] for i in sorted(tool_calls_data.keys())]
            if tool_calls_data
            else []
        )

        yield {
            "type": "done",
            "content": content.strip(),
            "tool_calls": tool_calls,
            "finish_reason": finish_reason,
        }

    def process_tool_call(
        self,
        tool_name: str,
        arguments: str,
        context: Optional[Dict[str, Any]] = None,
    ) -> Dict[str, Any]:
        """
        Process a tool call from the LLM

        Args:
            tool_name: Name of the tool to call
            arguments: JSON string of tool arguments

        Returns:
            Result of tool execution
        """
        try:
            args = json.loads(arguments) if arguments else {}
            if tool_name.startswith("shell_") and context:
                if "session_id" not in args and context.get("session_id"):
                    args["session_id"] = context["session_id"]
            result = tool_executor.execute_tool(tool_name, **args)
            return {"success": True, "result": result}
        except json.JSONDecodeError:
            return {
                "success": False,
                "error": f"Invalid JSON arguments: {arguments}",
            }
        except Exception as e:
            return {"success": False, "error": str(e)}


# Global agent instance
agent = Agent()



"""Tool executor and registry"""
from typing import Any, Dict, List
from .tools.base import BaseTool, BaseToolGroup, BaseStatus
from .tools.execute_command import ShellToolGroup, ShellStatus
from .tools.read_file import ReadFileTool
from .tools.write_file import WriteFileTool
from .tools.list_directory import ListDirectoryTool


class ToolExecutor:
    """Manages and executes tools"""
    
    def __init__(self):
        self.tools: Dict[str, BaseTool] = {}
        self.statuses: Dict[str, BaseStatus] = {}
        self._register_default_tools()
    
    def _register_default_tools(self) -> None:
        """Register all default tools"""
        default_tools = [
            ReadFileTool(),
            WriteFileTool(),
            ListDirectoryTool(),
        ]
        for tool in default_tools:
            self.register_tool(tool)

        self.register_tool_group(ShellToolGroup())
        self.register_status(ShellStatus())
    
    def register_tool(self, tool: BaseTool) -> None:
        """Register a new tool"""
        self.tools[tool.name] = tool

    def register_tool_group(self, group: BaseToolGroup) -> None:
        """Register a group of tools"""
        tools = group.get_tools()
        for tool in tools:
            self.register_tool(tool)

    def register_status(self, status: BaseStatus) -> None:
        """Register a status-only entry"""
        self.statuses[status.name] = status
    
    def execute_tool(self, tool_name: str, **kwargs) -> Dict[str, Any]:
        """Execute a registered tool"""
        if tool_name not in self.tools:
            return {
                "success": False,
                "error": f"Tool '{tool_name}' not found"
            }
        
        try:
            tool = self.tools[tool_name]
            result = tool.execute(**kwargs)
            return result
        except Exception as e:
            return {
                "success": False,
                "error": str(e)
            }
    
    def get_all_tools(self) -> List[Dict[str, Any]]:
        """Get all registered tools in OpenAI format"""
        return [tool.to_dict() for tool in self.tools.values()]

    def get_all_tool_statuses(self) -> List[Dict[str, Any]]:
        """Get status for all registered tools"""
        return [status.get_status() for status in self.statuses.values()]


# Global tool executor instance
tool_executor = ToolExecutor()



"""Base tool class and interfaces"""
from abc import ABC, abstractmethod
from typing import Any, Dict, List


class BaseTool(ABC):
    """Abstract base class for all tools"""
    
    @property
    @abstractmethod
    def name(self) -> str:
        """Tool name"""
        pass
    
    @property
    @abstractmethod
    def description(self) -> str:
        """Tool description"""
        pass
    
    @property
    @abstractmethod
    def parameters(self) -> Dict[str, Any]:
        """Tool parameters schema"""
        pass
    
    @abstractmethod
    def execute(self, **kwargs) -> Dict[str, Any]:
        """Execute the tool"""
        pass
    
    def to_dict(self) -> Dict[str, Any]:
        """Convert tool to OpenAI function calling format"""
        return {
            "type": "function",
            "function": {
                "name": self.name,
                "description": self.description,
                "parameters": self.parameters
            }
        }


class BaseToolGroup(ABC):
    """Abstract base class for tool groups"""

    @property
    @abstractmethod
    def name(self) -> str:
        """Tool group name"""
        pass

    @property
    def description(self) -> str:
        """Tool group description"""
        return ""

    @abstractmethod
    def get_tools(self) -> List[BaseTool]:
        """Return tools in this group"""
        pass

class BaseStatus(ABC):
    """Abstract base class for status-only display"""

    @property
    @abstractmethod
    def name(self) -> str:
        """Status name"""
        pass

    @property
    def description(self) -> str:
        """Status description"""
        return ""

    @abstractmethod
    def get_status(self) -> Dict[str, Any]:
        """Return status for UI/health checks"""
        pass



"""Main FastAPI application"""
import json
from pathlib import Path
from fastapi import FastAPI, HTTPException
from fastapi.staticfiles import StaticFiles
from fastapi.middleware.cors import CORSMiddleware
from fastapi.responses import StreamingResponse

from .config import settings
from .models import (
    SessionCreate,
    SystemPromptUpdate,
    ChatRequest,
    MessageRecord,
    ToolCallRecord,
    SystemPromptGenerateRequest,
    SystemPromptGenerateResponse,
)
from .session_manager import session_manager
from .agent import agent
from .executor import tool_executor
from .prompt_generator import generate_system_prompt

# Initialize FastAPI app
app = FastAPI(
    title=settings.app_title,
    debug=settings.debug
)

# Add CORS middleware
app.add_middleware(
    CORSMiddleware,
    allow_origins=settings.cors_origins,
    allow_credentials=True,
    allow_methods=["*"],
    allow_headers=["*"],
)


# Prompt Generation Endpoints
@app.post("/api/prompt/generate")
def generate_prompt(request: SystemPromptGenerateRequest):
    """
    Generate a system prompt with current date/time, safety policies, and tool descriptions
    
    Args:
        request: GeneratePromptRequest with optional custom_instructions
        
    Returns:
        Generated system prompt with metadata
    """
    from .prompt_generator import get_current_datetime_info
    
    prompt = generate_system_prompt(
        custom_instructions=request.custom_instructions
    )
    
    return SystemPromptGenerateResponse(
        system_prompt=prompt,
        datetime_info=get_current_datetime_info()
    )


@app.get("/api/prompt/generate")
def generate_prompt_get(custom_instructions: str = ""):
    """
    Generate a system prompt (GET endpoint for convenience)
    
    Query Parameters:
        custom_instructions: Optional custom instructions
        
    Returns:
        Generated system prompt with metadata
    """
    from .prompt_generator import get_current_datetime_info
    
    prompt = generate_system_prompt(
        custom_instructions=custom_instructions
    )
    
    return SystemPromptGenerateResponse(
        system_prompt=prompt,
        datetime_info=get_current_datetime_info()
    )


@app.post("/api/sessions")
def create_session(request: SessionCreate):
    """Create a new chat session"""
    # If system_prompt not provided, generate one automatically
    system_prompt = request.system_prompt
    if system_prompt is None:
        system_prompt = generate_system_prompt(custom_instructions="")
    
    session = session_manager.create_session(
        title=request.title,
        system_prompt=system_prompt
    )
    return session.model_dump()


@app.get("/api/sessions")
def list_sessions():
    """List all chat sessions"""
    sessions = session_manager.list_sessions()
    return [s.model_dump() for s in sessions]


@app.get("/api/sessions/{session_id}")
def get_session(session_id: str):
    """Get a specific session with all records"""
    session = session_manager.get_session(session_id)
    if not session:
        raise HTTPException(status_code=404, detail="Session not found")
    
    return session.model_dump()


@app.delete("/api/sessions/{session_id}")
def delete_session(session_id: str):
    """Delete a session"""
    success = session_manager.delete_session(session_id)
    if not success:
        raise HTTPException(status_code=404, detail="Session not found")
    return {"status": "success"}


@app.get("/api/tools/status")
def get_tools_status():
    """Get status of all tools"""
    return tool_executor.get_all_tool_statuses()


# Streaming chat endpoint
@app.post("/api/chat/stream")
def chat_stream(request: ChatRequest):
    """Send a message and get AI response with streaming support"""
    
    def generate():
        # Get session
        session = session_manager.get_session(request.session_id)
        if not session:
            yield f"data: {json.dumps({'type': 'error', 'content': 'Session not found'})}\n\n"
            return
        
        # Add user message record
        user_record = MessageRecord(
            role="user",
            content=request.message,
            binary_content=request.binary_content
        )
        session_manager.add_record(request.session_id, user_record.model_dump())
        
        # Send user record event
        yield f"data: {json.dumps({'type': 'record', 'record': user_record.model_dump()}, default=str)}\n\n"
        
        # Get messages for LLM
        messages = session_manager.get_messages_for_llm(request.session_id)
        
        try:
            # Tool calling loop
            max_iterations = 50
            
            for iteration in range(max_iterations):
                # Stream AI response
                current_content = ""
                tool_calls = []
                
                for chunk in agent.chat_stream(
                    messages=messages[1:],
                    system_prompt=session.system_prompt,
                    images=[bc.get("data") for bc in (request.binary_content or []) if bc.get("data")] if iteration == 0 else None
                ):
                    if chunk["type"] == "content":
                        current_content += chunk["content"]
                        yield f"data: {json.dumps({'type': 'content', 'content': chunk['content']})}\n\n"
                    elif chunk["type"] == "done":
                        tool_calls = chunk.get("tool_calls", [])
                
                if not tool_calls:
                    # No tool calls, save and finish
                    if current_content:
                        assistant_record = MessageRecord(
                            role="assistant",
                            content=current_content
                        )
                        session_manager.add_record(request.session_id, assistant_record.model_dump())
                        yield f"data: {json.dumps({'type': 'record', 'record': assistant_record.model_dump()}, default=str)}\n\n"
                    
                    yield f"data: {json.dumps({'type': 'done'})}\n\n"
                    return
                
                # Has tool calls - save thinking message if any
                if current_content:
                    thinking_record = MessageRecord(
                        role="assistant",
                        content=current_content
                    )
                    session_manager.add_record(request.session_id, thinking_record.model_dump())
                    yield f"data: {json.dumps({'type': 'record', 'record': thinking_record.model_dump()}, default=str)}\n\n"
                    messages.append({"role": "assistant", "content": current_content, "tool_calls": tool_calls})
                else:
                    messages.append({"role": "assistant", "content": "", "tool_calls": tool_calls})
                
                # Signal tool calls starting
                yield f"data: {json.dumps({'type': 'tool_calls_start', 'tool_calls': tool_calls})}\n\n"
                
                # Execute tool calls
                for tool_call in tool_calls:
                    tool_name = tool_call["function"]["name"]
                    tool_args_str = tool_call["function"]["arguments"]
                    tool_id = tool_call["id"]
                    
                    # Parse arguments
                    try:
                        tool_args = json.loads(tool_args_str) if isinstance(tool_args_str, str) else tool_args_str
                    except json.JSONDecodeError:
                        tool_args = {"raw_arguments": tool_args_str}
                    
                    # Signal tool execution starting
                    yield f"data: {json.dumps({'type': 'tool_start', 'tool_call_id': tool_id, 'tool_name': tool_name, 'arguments': tool_args, 'success': None})}\n\n"
                    
                    # Execute the tool
                    result = agent.process_tool_call(
                        tool_name,
                        tool_args_str,
                        context={"session_id": request.session_id}
                    )
                    
                    # Create and save tool call record
                    tool_record = ToolCallRecord(
                        tool_call_id=tool_id,
                        tool_name=tool_name,
                        arguments=tool_args,
                        result=result.get("result") if result.get("success") else result.get("error"),
                        success=result.get("success", False)
                    )
                    session_manager.add_record(request.session_id, tool_record.model_dump())
                    
                    # Send tool result
                    yield f"data: {json.dumps({'type': 'tool_result', 'record': tool_record.model_dump()}, default=str)}\n\n"
                    
                    # Add tool result to messages
                    messages.append({
                        "role": "tool",
                        "tool_call_id": tool_id,
                        "content": json.dumps(result) if isinstance(result, dict) else str(result)
                    })
            
            # Max iterations reached
            final_content = "Tool execution reached maximum iterations. Please try again with a simpler request."
            final_record = MessageRecord(role="assistant", content=final_content)
            session_manager.add_record(request.session_id, final_record.model_dump())
            yield f"data: {json.dumps({'type': 'record', 'record': final_record.model_dump()}, default=str)}\n\n"
            yield f"data: {json.dumps({'type': 'done'})}\n\n"
            
        except Exception as e:
            error_content = f"Error: {str(e)}"
            error_record = MessageRecord(role="assistant", content=error_content)
            session_manager.add_record(request.session_id, error_record.model_dump())
            yield f"data: {json.dumps({'type': 'error', 'content': error_content})}\n\n"
    
    return StreamingResponse(
        generate(),
        media_type="text/event-stream",
        headers={
            "Cache-Control": "no-cache",
            "Connection": "keep-alive",
            "X-Accel-Buffering": "no"
        }
    )


# System prompt endpoints
@app.post("/api/sessions/{session_id}/system-prompt")
def update_system_prompt(session_id: str, request: SystemPromptUpdate):
    """Update session system prompt"""
    session = session_manager.update_system_prompt(session_id, request.system_prompt)
    if not session:
        raise HTTPException(status_code=404, detail="Session not found")
    return session.model_dump()


# Health check
@app.get("/api/health")
def health_check():
    """Health check endpoint"""
    return {"status": "ok"}


# Serve frontend static files
static_dir = Path(__file__).parent / "static"
static_dir.mkdir(parents=True, exist_ok=True)
print(static_dir)
app.mount("/", StaticFiles(directory=static_dir, html=True), name="static")


if __name__ == "__main__":
    import uvicorn
    uvicorn.run(app, host="0.0.0.0", port=8009)



"""List directory tool"""
from pathlib import Path
from typing import Any, Dict
from .base import BaseTool


class ListDirectoryTool(BaseTool):
    """List files and directories"""
    
    @property
    def name(self) -> str:
        return "list_directory"
    
    @property
    def description(self) -> str:
        return "List files and directories in a path"
    
    @property
    def parameters(self) -> Dict[str, Any]:
        return {
            "type": "object",
            "properties": {
                "path": {
                    "type": "string",
                    "description": "Path to list"
                }
            },
            "required": ["path"]
        }
    
    def execute(self, path: str) -> Dict[str, Any]:
        """List directory contents"""
        try:
            items = []
            for item in Path(path).iterdir():
                items.append({
                    "name": item.name,
                    "type": "directory" if item.is_dir() else "file",
                    "path": str(item)
                })
            return {
                "success": True,
                "items": items
            }
        except Exception as e:
            return {
                "success": False,
                "error": str(e)
            }



"""System Prompt Generator for Agent"""

from datetime import datetime
from zoneinfo import ZoneInfo
from typing import List, Dict, Any
from pathlib import Path
from jinja2 import Environment, FileSystemLoader
from .executor import tool_executor


# Initialize Jinja2 Environment
TEMPLATE_DIR = Path(__file__).parent / "templates"
jinja_env = Environment(
    loader=FileSystemLoader(str(TEMPLATE_DIR)),
    trim_blocks=True,
    lstrip_blocks=True,
    autoescape=False
)


def get_current_datetime_info() -> Dict[str, str]:
    """
    Get current date, time and timezone information
    
    Returns:
        Dictionary containing formatted date/time/timezone
    """
    # Get current UTC time
    utc_now = datetime.now(ZoneInfo("UTC"))
    
    # Also get local timezone (system timezone)
    local_tz = datetime.now().astimezone().tzinfo
    local_now = datetime.now(local_tz)
    
    return {
        "current_utc": utc_now.strftime("%Y-%m-%d %H:%M:%S UTC"),
        "current_local": local_now.strftime("%Y-%m-%d %H:%M:%S %Z"),
        "timezone": str(local_tz),
    }


def get_available_tools_description() -> str:
    """
    Get descriptions of all available tools for the prompt
    
    Returns:
        Formatted string describing all available tools
    """
    tools = tool_executor.get_all_tools()
    
    if not tools:
        return "No tools are currently available."
    
    tools_description = "## Available Tools\n\n"
    
    for tool in tools:
        # Extract tool info from OpenAI function format
        func = tool.get('function', {})
        tool_name = func.get('name', 'Unknown')
        tool_desc = func.get('description', 'No description available')
        params = func.get('parameters', {})
        
        tools_description += f"### {tool_name}\n"
        tools_description += f"Description: {tool_desc}\n"
        
        if 'properties' in params:
            tools_description += "Parameters:\n"
            for param_name, param_info in params['properties'].items():
                param_type = param_info.get('type', 'unknown')
                param_desc = param_info.get('description', 'No description')
                required = param_name in params.get('required', [])
                req_mark = " (required)" if required else " (optional)"
                tools_description += f"  - `{param_name}` ({param_type}){req_mark}: {param_desc}\n"
        
        tools_description += "\n"
    
    return tools_description


def generate_system_prompt(custom_instructions: str = "") -> str:
    """
    Generate a system prompt for the Agent using Jinja2 template
    
    Args:
        custom_instructions: Optional custom instructions to append to the prompt
    
    Returns:
        Complete system prompt string
    """
    datetime_info = get_current_datetime_info()
    tools_desc = get_available_tools_description()
    
    # Load and render template
    template = jinja_env.get_template("system_prompt.j2")
    
    prompt = template.render(
        current_utc=datetime_info['current_utc'],
        current_local=datetime_info['current_local'],
        timezone=datetime_info['timezone'],
        tools_description=tools_desc,
        custom_instructions=custom_instructions
    )
    
    return prompt
